#ifndef BOOZ_VECTOR_MATRIX_H
#define BOOZ_VECTOR_MATRIX_H

#include <Eigen/Core>

namespace booz_xform {

  typedef double boozfloat;
  
  typedef Eigen::ArrayXd Vector;
  typedef Eigen::ArrayXi IntVector;
  typedef Eigen::ArrayXXd Matrix;

} // namespace booz_xform


#endif

